#main_window
{
  background: qlineargradient(x1: 0, y1: 0, x2:0,y2:1, stop: 0 white , stop: 1 #E8E8E8);
}

MiniWindow QPushButton {
  color: #777;
  border: 1px solid #CCC;
  border-radius: 0px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 white, stop: 1 #E6E6E6);
  min-height: 30px;
  min-width: 30px;
}

#send_button{
  color: #FFF;
  border: 1px solid #3786E6;
  border-radius: 4px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 #72B2F8, stop: 1 #3484E6);
  padding: 2px;
  width: 20px;
}

#send_button:disabled{
  color: #D3E8FE;
  border: 1px solid #6DAEF7;
  border-radius: 4px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 #A5CFFA, stop: 1 #72B2F8);
}
#address_input, #amount_input, #label_input
{
    color: #000;
    padding: 5px;
    border-radius: 5px;
    min-height: 23px;
    border: 1px solid #AAA9A9;
    width: 200px;
}

#address_input[isValid=true]
{
  color: #4D9948;
}

#address_input[isValid=false]
{
  color: #CE4141;
}

#balance_label
{
    color: #333;
}

#history
{
    color: #888;
}

