ELECTRUM_VERSION = "1.9.8"  # version of the client package
PROTOCOL_VERSION = '0.9'    # protocol version requested
NEW_SEED_VERSION = 6        # bip32 wallets
OLD_SEED_VERSION = 4        # old electrum deterministic generation
SEED_PREFIX      = '01'     # the hash of the mnemonic seed must begin with this
