#main_window
{
  background: qlineargradient(x1: 0, y1: 0, x2:0,y2:1, stop: 0 white , stop: 1 #F2E3BE);
}

MiniWindow QPushButton {
  color: #C1A76D;
  border: 1px solid #A7811C;
  border-radius: 0px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 white, stop: 1 #F2E3BE);
  min-height: 25px;
  min-width: 30px;
}

#send_button{
  color: #FEEBA7;
  border: 1px solid #AD8B35;
  border-radius: 4px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 #E0A035, stop: 1 #AD8B35);
  min-width: 80px;
  min-height: 23px;
  padding: 2px;
}

#send_button:disabled{
  color: #FEEDD3;
  border: 1px solid #F7D46D;
  border-radius: 4px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 #FAEEA5, stop: 1 #DBC050);
  min-width: 80px;
  min-height: 23px;
  padding: 2px;
}

#receive_button
{
  color: #FEEBA7;
  border: 1px solid #AD8B35;
  border-radius: 4px;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 #E0A035, stop: 1 #AD8B35);
  min-height: 25px;
  min-width: 30px;
}          
#receive_button[isActive=true]
{
  color: #FEEBA7;
  background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
  stop: 0 #E0A035, stop: 1 #987620);
} 

#address_input, #amount_input
{
    color: #000;
    padding: 5px;
    border-radius: 4px;
    border: 1px solid #CBAE69;
    width: 225px;
}

#address_input[isValid=true]
{
  color: #4D9948;
  padding: 5px;
  border-radius: 4px;
  border: 1px solid #CBAE69;
  width: 225px;
  margin-top: 4px;
}

#address_input[isValid=false]
{
  color: #CE4141;
  padding: 5px;
  border-radius: 4px;
  border: 1px solid #CBAE69;
  width: 225px;
  margin-top: 4px;
}

#address_input[isValid=placeholder]
{
  color: #DEC58D;
  padding: 5px;
  border-radius: 4px;
  border: 1px solid #CBAE69;
  width: 225px;
  margin-top: 4px;
}
#balance_label
{
    color: #7E5907;
}

#history
{
    color: #8B6914;
}

